"""Standalone script to produce Gantt Charts from AutoSubmit logs.

Example usage:

    $ python create_ganttchart.py o005 --chunks 38 52 --ignore AQUA

This produces the figure posted in:

    https://earth.bsc.es/gitlab/digital-twins/de_340-2/simulation-runs/-/issues/82#note_368241

Full options for plot creation:

    $ python create_ganttchart.py -h
"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter
from datetime import datetime
import warnings
from pathlib import Path
import tempfile

import matplotlib.pyplot as plt
import pandas as pd

from destine_analytics.logreport import get_aslogs_df, jobs_df_from_aslogs_df
from destine_analytics.plots import (
    plot_job_gantt_chart as inner_plot_job_gantt_chart,
    plot_job_gantt_chart_plotly,
)
from destine_analytics.decorators.asrun import (
    asrun_with_jobs_available,
    asrun_with_stats,
)
from destine_analytics.paths import LOG_DIR

from .utils import sanitize_output, transfer_logs_from_remote

# Prefixes of files to ignore in Gantt Chart by default
IGNORE_PREFIX = (".", "ls", "config", "githook", "mother_request", "only_lra")


def _get_parser() -> ArgumentParser:
    """Configures and returns the argument parser for the script."""
    parser = ArgumentParser(
        description="Gantt Chart plotter for Autosubmit logs.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "expid",
        type=str,
        help="Experiment id.",
    )
    parser.add_argument(
        "-e",
        "--ssh-command",
        nargs="?",
        const="__use_env__",
        default=None,
        help=(
            "Trigger remote log transfer. "
            "If an SSH command is provided as an argument, it will be used. "
            "Otherwise, credentials from the .env file will be used."
        ),
    )
    parser.add_argument(
        "--chunks",
        type=int,
        nargs=2,
        metavar=("START", "END"),
        default=(None, None),
        help="First and last chunks to plot.",
    )
    parser.add_argument(
        "--ignore",
        type=str,
        nargs="*",
        default=[],
        help="Ignore jobs starting with these arguments.",
    )
    parser.add_argument(
        "--color",
        choices=("job", "chunk", "status"),
        default="chunk",
        help="Whether to color jobs by job, by chunk or by status.",
    )
    parser.add_argument(
        "--starttype",
        choices=("started", "submitted"),
        default="started",
        help="Whether the start of the bars is at AS 'started' or 'submitted' status.",
    )
    parser.add_argument(
        "--cmap", type=str, default="tab20", help="Matplotlib colormap to use."
    )
    parser.add_argument(
        "--interval",
        type=int,
        default=-1,
        help="Interval for time ticks in x axis. -1 sets it to 'automatic'. Default: -1",
    )
    parser.add_argument(
        "--pivot",
        action="store_true",
        help="All bars are classified by chunk in the y axis.",
    )
    parser.add_argument(
        "--no-pivot",
        action="store_false",
        dest="autopivot",
        help="Prevent bars from being classified by chunk in the y axis, even if splits overlap.",
    )
    parser.add_argument(
        "--by-split",
        action="store_true",
        help="Gantt Chart entries are separated by split.",
    )
    parser.add_argument(
        "--by-member",
        action="store_true",
        help="Gantt Chart entries are separated by member.",
    )
    parser.add_argument(
        "--durations",
        action="store_true",
        help="Add the summed durations of all the jobs of each entry in the Gantt chart.",
    )
    parser.add_argument(
        "--no-edge",
        action="store_false",
        dest="edge",
        help="Removes the black edge from the bars in the Gantt chart.",
    )
    parser.add_argument(
        "--dpi",
        type=int,
        help="DPI for the saved figure.",
    )
    parser.add_argument(
        "--after",
        type=str,
        help="Date after which jobs are read. Format: YYYYMMDD or YYYYMMDDhhmmss",
    )
    parser.add_argument(
        "--y-pos",
        type=float,
        nargs="+",
        dest="y_positions",
        help="Relative positions of line info",
    )
    parser.add_argument(
        "--y-bar",
        type=float,
        default=0.4,
        dest="bar_height",
        help="Height of the bars. Default: 0.4",
    )
    parser.add_argument(
        "--y-factor",
        type=float,
        default=0.4,
        dest="height_factor",
        help="Factor to create height of the plot. Default: 0.4",
    )
    parser.add_argument(
        "-A",
        "--add-information",
        type=str,
        choices=("available", "stats", None),
        default=None,
        help="Extra info to add to the Gantt Chart.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Enable verbose output, especially for remote operations.",
    )
    parser.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        help="Show the Gantt Chart in an interactive window.",
    )
    parser.add_argument(
        "-o",
        "--output",
        nargs="?",
        type=str,
        default=None,
        const="__default__",
        help=(
            "Output directory or file for saving figures. If no path is given, "
            "defaults to 'figures/'. If flag is not used, figure is not saved."
        ),
    )
    parser.set_defaults(autopivot=True)
    return parser


def _filter_by_chunk_interval(df: pd.DataFrame, chunks: tuple) -> pd.DataFrame:
    """Filters a DataFrame to include only entries within a specified chunk range."""
    return df[(df["chunk"] >= chunks[0]) & (df["chunk"] <= chunks[1])]


def _parse_date_or_none(after: str | None) -> datetime | None:
    """Parses a date string into a datetime object."""
    if after is None:
        return None
    if len(after) == 8:
        return datetime.strptime(after, r"%Y%m%d")
    if len(after) == 14:
        return datetime.strptime(after, r"%Y%m%d%H%M%S")
    raise ValueError("Date must be in format YYYYMMDD or YYYYYMMDDhhmmss.")


def _decorate_plot_job_gantt_chart(add_information, aslogs_path, args):
    """
    Decorates the plotting function with additional functionality based on arguments.

    This function wraps the core plotting function with decorators to add features
    like statistics or job availability information.
    """
    if add_information is None:
        return inner_plot_job_gantt_chart
    if add_information == "stats":
        return asrun_with_stats(
            inner_plot_job_gantt_chart,
            log_path=aslogs_path,
            after=args.after,
            rel_y_position=args.y_positions,
            color_text="#006600",
        )
    if add_information == "available":
        return asrun_with_jobs_available(
            inner_plot_job_gantt_chart,
            log_path=aslogs_path,
            after=args.after,
            rel_y_position=args.y_positions,
        )
    raise ValueError(f"Invalid add_information: {add_information}")


def _filter_jobs(jobs: pd.DataFrame, args) -> tuple[pd.DataFrame, str, str]:
    """
    Filters the jobs DataFrame based on command-line arguments.

    Args:
        jobs: DataFrame containing job information.
        args: Parsed command-line arguments.

    Returns:
        A tuple containing:
        - The filtered jobs DataFrame.
        - The title for the plot.
        - The recommended filename for the plot.
    """
    chunks0, chunks1 = args.chunks
    ignore = args.ignore
    after = _parse_date_or_none(args.after)
    starttype = args.starttype
    expid = args.expid

    # Filter jobs by chunk interval
    if chunks0 is not None and chunks1 is not None:
        jobs = _filter_by_chunk_interval(jobs, (chunks0, chunks1))
        title = f"{expid} Simulation Timeline - Chunks [{chunks0}, {chunks1}]"
        filename = f"{expid}_simulation_timeline_{chunks0}_{chunks1}.png"
    else:
        title = f"{expid} Simulation Timeline"
        filename = f"{expid}_simulation_timeline.png"

    # Filter jobs by ignore prefix
    jobs = jobs[
        jobs["job"].apply(
            lambda x: not any(x.startswith(element) for element in ignore)
        )
    ]

    # Filter jobs by start date
    if after is not None:
        jobs = jobs[jobs[starttype] >= after]

    return jobs, title, filename


def makefigure(jobs: pd.DataFrame, aslogs_path, args, title: str):
    """
    Creates the Gantt chart figure using either Matplotlib or Plotly.

    Args:
        jobs: The DataFrame with job data to plot.
        aslogs_path: Path to the ASLOGS directory, used by decorators.
        args: Parsed command-line arguments.
        title: The title for the figure.
    """
    pivot_options = (
        "member"
        if args.by_member
        else (
            "split"
            if args.by_split
            else "chunk" if args.pivot else "on_overlap" if args.autopivot else False
        )
    )

    print("Plotting Gantt chart...")
    if args.interactive:
        # Create an interactive Plotly figure
        fig = plot_job_gantt_chart_plotly(
            jobs,
            title=title,
            color_key=args.color,
            starttype=args.starttype,
            pivot=pivot_options,
            add_durations=args.durations,
            height_factor=args.height_factor,
        )
        print("Showing interactive Gantt chart.")
        fig.show(
            config={
                "scrollZoom": True,
                "displayModeBar": True,
                "modeBarButtonsToRemove": ["pan2d", "lasso2d", "select2d"],
                "displaylogo": False,
                "responsive": True,
            }
        )
    else:
        # Create a static Matplotlib figure
        plot_function = _decorate_plot_job_gantt_chart(
            args.add_information,
            aslogs_path,
            args,
        )
        cmap = plt.get_cmap(args.cmap)
        fig = plot_function(
            jobs,
            title=title,
            color_key=args.color,
            starttype=args.starttype,
            cmap=cmap,
            interval=args.interval,
            pivot=pivot_options,
            add_durations=args.durations,
            edge=args.edge,
            bar_height=args.bar_height,
            height_factor=args.height_factor,
        )

    return fig


def _inner(args, experiment_path: Path):
    """
    The main execution logic of the script after paths are resolved.

    Args:
        args: Parsed command-line arguments.
        experiment_path: Path to the root of the experiment data (local or temporary).
    """
    print("Reading logs...")
    aslogs_path = experiment_path / "ASLOGS"
    asfiles_path = experiment_path / "ASFILES"
    aslogs = get_aslogs_df(asfiles_path, ignore_prefix=IGNORE_PREFIX)

    print("Extracting jobs...")
    jobs = jobs_df_from_aslogs_df(aslogs, asfiles_path)

    # Filter jobs and generate plot title/filename
    jobs, title, filename = _filter_jobs(jobs, args)

    # Create the figure
    fig = makefigure(jobs, aslogs_path, args, title)

    # Save the figure if an output path is provided
    save_path = sanitize_output(args.output, filename)
    if save_path:
        save_path.parent.mkdir(parents=True, exist_ok=True)
        fig.savefig(save_path, dpi=args.dpi)
        print(f"Saved {save_path}")


def main():
    """
    Parses arguments and orchestrates the Gantt chart creation.

    This function handles argument parsing, validation, and determines whether
    to run in local or remote mode.
    """
    # Parse command-line arguments
    args = _get_parser().parse_args()

    # 1. Argument validation
    if not args.interactive and not args.output:
        raise ValueError("Interactive mode is required when output is not provided.")

    if args.by_member and not args.by_split:
        warnings.warn(
            "Independent pivots are not yet implemented. "
            "Therefore, --by-member has the same effect as --by-member --by-split."
        )

    if args.add_information and args.interactive:
        raise NotImplementedError(
            "Adding information is not supported for interactive plots yet."
        )

    # 2. Execution Logic
    # Determine whether to use local data or transfer from a remote machine
    if args.ssh_command is not None:
        # Remote mode: ssh_command is either a string or the const.
        ssh_cmd_to_pass = (
            args.ssh_command if args.ssh_command != "__use_env__" else None
        )

        # Give user feedback
        if ssh_cmd_to_pass:
            print(f"SSH command provided. Running in remote mode for expid '{args.expid}'.")
        else:
            print(
                f"SSH flag provided without command. Using .env for remote credentials for expid '{args.expid}'."
            )

        with tempfile.TemporaryDirectory() as temp_dir:
            log_dir = Path(temp_dir)
            experiment_path = transfer_logs_from_remote(
                args.expid,
                log_dir,
                ssh_command=ssh_cmd_to_pass,
                verbose=args.verbose,
            )
            _inner(args, experiment_path)
    else:
        # Local mode: use logs from the configured LOG_DIR
        print(f"No SSH command. Running in local mode for expid '{args.expid}'.")
        log_dir = LOG_DIR
        experiment_path = log_dir / args.expid
        _inner(args, experiment_path)


if __name__ == "__main__":
    main()
